@@ -1,5 +1,8 @@
 #' Apply a Function over a List or Vector via Futures
 #'
+#' `future_lapply()` implements [base::lapply()] using futures, and
+#' analogously for all the other `future_nnn()` functions.
+#' 
 #' @param X  A vector-like object to iterate over.
 #' 
 #' @param FUN  A function taking at least one argument.
@@ -29,7 +32,11 @@
 #'        (if there are enough elements in `X`).
 #'        If `Inf` or `FALSE`, then one future per element of
 #'        `X` is used.
+#'        Only used if `future.chunk.size` is `NULL`.
 #'
+#' @param future.chunk.size The average number of elements per future ("chunk").
+#'        If `NULL`, then argument `future.scheduling` is used.
+#' 
 #' @return
 #' For `future_lapply()`, a list with same length and names as `X`.
 #' See [base::lapply()] for details.
@@ -89,14 +96,11 @@
 #'
 #' @keywords manip programming iteration
 #'
-#' @importFrom globals globalsByName cleanup
+#' @importFrom globals globalsByName
 #' @importFrom future future resolve values as.FutureGlobals nbrOfWorkers getGlobalsAndPackages FutureError
-#' @importFrom parallel nextRNGStream nextRNGSubStream splitIndices
 #' @importFrom utils capture.output head str
 #' @export
-future_lapply <- function(X, FUN, ..., future.globals = TRUE, future.packages = NULL, future.seed = FALSE, future.lazy = FALSE, future.scheduling = 1.0) {
-  objectSize <- import_future("objectSize")
-  
+future_lapply <- function(X, FUN, ..., future.globals = TRUE, future.packages = NULL, future.seed = FALSE, future.lazy = FALSE, future.scheduling = 1.0, future.chunk.size = NULL) {
   stop_if_not(is.function(FUN))
   
   stop_if_not(is.logical(future.lazy))
@@ -124,207 +128,38 @@ future_lapply <- function(X, FUN, ..., future.globals = TRUE, future.packages =
   envir <- future.envir
   
   ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  ## 1. Global variables
+  ## 1. Globals and Packages
   ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  ## The default is to gather globals
-  if (is.null(future.globals)) future.globals <- TRUE
-
-  packages <- NULL
-  globals <- future.globals
-  if (is.logical(globals)) {
-    ## Gather all globals?
-    if (globals) {
-      if (debug) mdebug("Finding globals ...")
-
-      expr <- do.call(call, args = c(list("FUN"), list(...)))
-      gp <- getGlobalsAndPackages(expr, envir = envir, globals = TRUE)
-      globals <- gp$globals
-      packages <- gp$packages
-      gp <- NULL
-      
-      if (debug) {
-        mdebug(" - globals found: [%d] %s", length(globals), hpaste(sQuote(names(globals))))
-        mdebug(" - needed namespaces: [%d] %s", length(packages), hpaste(sQuote(packages)))
-        mdebug("Finding globals ... DONE")
-      }
-    } else {
-      ## globals = FALSE
-      globals <- c("FUN", names(list(...)), "...")
-      globals <- globalsByName(globals, envir = envir, mustExist = FALSE)
-    }
-  } else if (is.character(globals)) {
-    globals <- unique(c(globals, "FUN", names(list(...)), "..."))
-    globals <- globalsByName(globals, envir = envir, mustExist = FALSE)
-  } else if (is.list(globals)) {
-    names <- names(globals)
-    if (length(globals) > 0 && is.null(names)) {
-      stop("Invalid argument 'future.globals'. All globals must be named.")
-    }
-  } else {
-    stop("Invalid argument 'future.globals': ", mode(globals))
-  }
-  globals <- as.FutureGlobals(globals)
-  stop_if_not(inherits(globals, "FutureGlobals"))
-  
-  names <- names(globals)
-  if (!is.element("FUN", names)) {
-    globals <- c(globals, FUN = FUN)
-  }
-  
-  if (!is.element("...", names)) {
-    if (debug) mdebug("Getting '...' globals ...")
-    dotdotdot <- globalsByName("...", envir = envir, mustExist = TRUE)
-    dotdotdot <- as.FutureGlobals(dotdotdot)
-    dotdotdot <- resolve(dotdotdot)
-    attr(dotdotdot, "total_size") <- objectSize(dotdotdot)
-    if (debug) mdebug("Getting '...' globals ... DONE")
-    globals <- c(globals, dotdotdot)
-  }
-
-  ## Assert there are no reserved variables names among globals
-  reserved <- intersect(c("...future.FUN", "...future.X_ii",
-                          "...future.seeds_ii"), names)
-  if (length(reserved) > 0) {
-    stop("Detected globals using reserved variables names: ",
-         paste(sQuote(reserved), collapse = ", "))
-  }
- 
-  ## Avoid FUN() clash with lapply(..., FUN) below.
-  names <- names(globals)
-  names[names == "FUN"] <- "...future.FUN"
-  names(globals) <- names
-
-  if (debug) {
-    mdebug("Globals to be used in all futures:")
-    mdebug(paste(capture.output(str(globals)), collapse = "\n"))
-  }
-
-
-  ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  ## 2. Packages
-  ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  if (!is.null(future.packages)) {
-    stop_if_not(is.character(future.packages))
-    future.packages <- unique(future.packages)
-    stop_if_not(!anyNA(future.packages), all(nzchar(future.packages)))
-    packages <- unique(c(packages, future.packages))
-  }
-  
-  if (debug) {
-    mdebug("Packages to be attached in all futures:")
-    mdebug(paste(capture.output(str(packages)), collapse = "\n"))
-  }
-
+  gp <- getGlobalsAndPackagesXApply(FUN = FUN,
+                                    args = list(...),
+                                    envir = envir,
+                                    future.globals = future.globals,
+                                    future.packages = future.packages,
+                                    debug = debug)
+  packages <- gp$packages
+  globals <- gp$globals
+  scanForGlobals <- gp$scanForGlobals
+  gp <- NULL
 
   ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   ## 3. Reproducible RNG (for sequential and parallel processing)
   ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  seed <- future.seed
+  seeds <- make_rng_seeds(nX, seed = future.seed, debug = debug)
 
-  ## Placeholder for all RNG stream seeds.
-  seeds <- NULL
-  
-  ## Don't use RNGs? (seed = FALSE)
-  if (is.logical(seed) && !is.na(seed) && !seed) seed <- NULL
-
-  # Use RNGs?
-  if (!is.null(seed)) {
-    if (debug) mdebug("Generating random seeds ...")
-
-    ## future_lapply() should return with the same RNG state regardless of
-    ## future strategy used. This is be done such that RNG kind is preserved
-    ## and the seed is "forwarded" one step from what it was when this
-    ## function was called. The forwarding is done by generating one random
-    ## number. Note that this approach is also independent on length(X) and
-    ## the diffent FUN() calls.
+  ## If RNG seeds are used (given or generated), make sure to reset
+  ## the RNG state afterward
+  if (!is.null(seeds)) {
     oseed <- next_random_seed()
-    on.exit(set_random_seed(oseed))    
-
-    ## A pregenerated sequence of random seeds?
-    if (is.list(seed)) {
-      if (debug) mdebug("Using a pre-define stream of random seeds ...", nX)
-      
-      nseed <- length(seed)
-      if (nseed != nX) {
-        stop("Argument 'seed' is a list, which specifies the sequence of seeds to be used for each element in 'X', but length(seed) != length(X): ", nseed, " != ", nX)
-      }
-
-      ## Assert same type of RNG seeds?
-      ns <- unique(unlist(lapply(seed, FUN = length), use.names = FALSE))
-      if (length(ns) != 1) {
-        stop("The elements of the list specified in argument 'seed' are not all of the same lengths (did you really pass RNG seeds?): ", hpaste(ns))
-      }
-
-      ## Did use specify scalar integers as meant for set.seed()?
-      if (ns == 1L) {
-        stop("Argument 'seed' is invalid. Pre-generated random seeds must be valid .Random.seed seeds, which means they should be all integers and consists of two or more elements, not just one.")
-      }
-
-      types <- unlist(lapply(seed, FUN = typeof), use.names = FALSE)
-      if (!all(types == "integer")) {
-        stop("The elements of the list specified in argument 'seed' are not all integers (did you really pass RNG seeds?): ", hpaste(unique(types)))
-      }
-      
-      ## Check if valid random seeds are specified.
-      ## For efficiency, only look at the first one.
-      if (!is_valid_random_seed(seed[[1]])) {
-        stop("The list in argument 'seed' does not seem to hold elements that are valid .Random.seed values: ", capture.output(str(seeds[[1]])))
-      }
-
-      seeds <- seed
-      
-      if (debug) mdebug("Using a pre-define stream of random seeds ... DONE", nX)
-    } else {
-      if (debug) mdebug("Generating random seed streams for %d elements ...", nX)
-      
-      ## Generate sequence of _all_ RNG seeds starting with an initial seed
-      ## '.seed' that is based on argument 'seed'.
-      .seed <- as_lecyer_cmrg_seed(seed)
-
-      seeds <- vector("list", length = nX)
-      for (ii in seq_len(nX)) {
-        ## RNG substream seed used in call FUN(X[[ii]], ...):
-        ## This way each future can in turn generate further seeds, also
-        ## recursively, with minimal risk of generating the same seeds as
-        ## another future. This should make it safe to recursively call
-        ## future_lapply(). /HB 2017-01-11
-        seeds[[ii]] <- nextRNGSubStream(.seed)
-        
-        ## Main random seed for next iteration (= ii + 1)
-        .seed <- nextRNGStream(.seed)
-      }
+    on.exit(set_random_seed(oseed))
+  }
   
-      if (debug) mdebug("Generating random seed streams for %d elements ... DONE", nX)
-    }
-    
-    if (debug) mdebug("Generating random seeds ... DONE")
-  } ## if (!is.null(seed))
-
   
   ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   ## 4. Load balancing ("chunking")
   ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  if (is.logical(future.scheduling)) {
-    if (future.scheduling) {
-      nbr_of_futures <- nbrOfWorkers()
-      if (nbr_of_futures > nX) nbr_of_futures <- nX
-    } else {
-      nbr_of_futures <- nX
-    }
-  } else {
-    ## Treat 'future.scheduling' as the number of futures per worker.
-    stop_if_not(future.scheduling >= 0)
-    nbr_of_workers <- nbrOfWorkers()
-    if (nbr_of_workers > nX) nbr_of_workers <- nX
-    nbr_of_futures <- future.scheduling * nbr_of_workers
-    if (nbr_of_futures < 1) {
-      nbr_of_futures <- 1L
-    } else if (nbr_of_futures > nX) {
-      nbr_of_futures <- nX
-    }
-  }
-
-  chunks <- splitIndices(nX, ncl = nbr_of_futures)
+  chunks <- makeChunks(nX, nbrOfWorkers = nbrOfWorkers(),
+                       future.scheduling = future.scheduling,
+                       future.chunk.size = future.chunk.size)
   if (debug) mdebug("Number of chunks: %d", length(chunks))   
 
   
@@ -332,17 +167,27 @@ future_lapply <- function(X, FUN, ..., future.globals = TRUE, future.packages =
   ## 5. Create futures
   ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   ## Add argument placeholders
-  globals_extra <- as.FutureGlobals(list(...future.X_ii = NULL, ...future.seeds_ii = NULL))
+  globals_extra <- as.FutureGlobals(list(
+    ...future.elements_ii = NULL,
+    ...future.seeds_ii = NULL,
+    ...future.globals.maxSize = NULL
+  ))
   attr(globals_extra, "resolved") <- TRUE
   attr(globals_extra, "total_size") <- 0
   globals <- c(globals, globals_extra)
 
+  
   ## At this point a globals should be resolved and we should know their total size
 ##  stop_if_not(attr(globals, "resolved"), !is.na(attr(globals, "total_size")))
 
-    ## To please R CMD check
-  ...future.FUN <- ...future.X_ii <- ...future.seeds_ii <- NULL
+  ## To please R CMD check
+  ...future.FUN <- ...future.elements_ii <- ...future.seeds_ii <-
+                   ...future.globals.maxSize <- NULL
 
+  globals.maxSize <- getOption("future.globals.maxSize")
+  globals.maxSize.default <- globals.maxSize
+  if (is.null(globals.maxSize.default)) globals.maxSize.default <- 500 * 1024^2
+  
   nchunks <- length(chunks)
   fs <- vector("list", length = nchunks)
   if (debug) mdebug("Number of futures (= number of chunks): %d", nchunks)
@@ -352,30 +197,87 @@ future_lapply <- function(X, FUN, ..., future.globals = TRUE, future.packages =
     chunk <- chunks[[ii]]
     if (debug) mdebug("Chunk #%d of %d ...", ii, length(chunks))
 
-    ## Subsetting outside future is more efficient
+    X_ii <- X[chunk]
     globals_ii <- globals
-    globals_ii[["...future.X_ii"]] <- X[chunk]
+    ## Subsetting outside future is more efficient
+    globals_ii[["...future.elements_ii"]] <- X_ii
+    packages_ii <- packages
+
+    if (scanForGlobals) {
+      mdebug(" - Finding globals in 'X' for chunk #%d ...", ii)
+      ## Search for globals in 'X_ii':
+      gp <- getGlobalsAndPackages(X_ii, envir = envir, globals = TRUE)
+      globals_X <- gp$globals
+      packages_X <- gp$packages
+      gp <- NULL
+
+      if (debug) {
+        mdebug("   + globals found in 'X' for chunk #%d: [%d] %s", chunk, length(globals_X), hpaste(sQuote(names(globals_X))))
+        mdebug("   + needed namespaces for 'X' for chunk #%d: [%d] %s", chunk, length(packages_X), hpaste(sQuote(packages_X)))
+      }
+    
+      ## Export also globals found in 'X_ii'
+      if (length(globals_X) > 0L) {
+        reserved <- intersect(c("...future.FUN", "...future.elements_ii",
+                                "...future.seeds_ii"), names(globals_X))
+        if (length(reserved) > 0) {
+          stop("Detected globals in 'X' using reserved variables names: ",
+               paste(sQuote(reserved), collapse = ", "))
+        }
+        globals_X <- as.FutureGlobals(globals_X)
+        globals_ii <- unique(c(globals_ii, globals_X))
+
+        ## Packages needed due to globals in 'X_ii'?
+        if (length(packages_X) > 0L)
+          packages_ii <- unique(c(packages_ii, packages_X))
+      }
+      mdebug(" - Finding globals in 'X' for chunk #%d ... DONE", ii)
+    }
+    
+    X_ii <- NULL
 ##    stop_if_not(attr(globals_ii, "resolved"))
+
+    ## Adjust option 'future.globals.maxSize' to account for the fact that more
+    ## than one element is processed per future.  The adjustment is done by
+    ## scaling up the limit by the number of elements in the chunk.  This is
+    ## a "good enough" approach.
+    ## (https://github.com/HenrikBengtsson/future.apply/issues/8).
+    if (length(chunk) > 1L) {
+      globals_ii["...future.globals.maxSize"] <- list(globals.maxSize)
+      options(future.globals.maxSize = length(chunk) * globals.maxSize.default)
+      if (debug) mdebug(" - Adjusted option 'future.globals.maxSize': %g -> %d * %g = %g (bytes)", globals.maxSize.default, length(chunk), globals.maxSize.default, getOption("future.globals.maxSize"))
+      on.exit(options(future.globals.maxSize = globals.maxSize), add = TRUE)
+    }
     
     ## Using RNG seeds or not?
     if (is.null(seeds)) {
       if (debug) mdebug(" - seeds: <none>")
       fs[[ii]] <- future({
-        lapply(seq_along(...future.X_ii), FUN = function(jj) {
-           ...future.X_jj <- ...future.X_ii[[jj]]
+        ...future.globals.maxSize.org <- getOption("future.globals.maxSize")
+        if (!identical(...future.globals.maxSize.org, ...future.globals.maxSize)) {
+          oopts <- options(future.globals.maxSize = ...future.globals.maxSize)
+          on.exit(options(oopts), add = TRUE)
+        }
+        lapply(seq_along(...future.elements_ii), FUN = function(jj) {
+           ...future.X_jj <- ...future.elements_ii[[jj]]
            ...future.FUN(...future.X_jj, ...)
         })
-      }, envir = envir, lazy = future.lazy, globals = globals_ii, packages = packages)
+      }, envir = envir, lazy = future.lazy, globals = globals_ii, packages = packages_ii)
     } else {
       if (debug) mdebug(" - seeds: [%d] <seeds>", length(chunk))
       globals_ii[["...future.seeds_ii"]] <- seeds[chunk]
       fs[[ii]] <- future({
-        lapply(seq_along(...future.X_ii), FUN = function(jj) {
-           ...future.X_jj <- ...future.X_ii[[jj]]
+        ...future.globals.maxSize.org <- getOption("future.globals.maxSize")
+        if (!identical(...future.globals.maxSize.org, ...future.globals.maxSize)) {
+          oopts <- options(future.globals.maxSize = ...future.globals.maxSize)
+          on.exit(options(oopts), add = TRUE)
+        }
+        lapply(seq_along(...future.elements_ii), FUN = function(jj) {
+           ...future.X_jj <- ...future.elements_ii[[jj]]
            assign(".Random.seed", ...future.seeds_ii[[jj]], envir = globalenv(), inherits = FALSE)
            ...future.FUN(...future.X_jj, ...)
         })
-      }, envir = envir, lazy = future.lazy, globals = globals_ii, packages = packages)
+      }, envir = envir, lazy = future.lazy, globals = globals_ii, packages = packages_ii)
     }
     
     ## Not needed anymore
@@ -412,27 +314,7 @@ future_lapply <- function(X, FUN, ..., future.globals = TRUE, future.packages =
     mdebug(" - Number of values expected: %d", nX)
   }
 
-  if (length(values2) != nX) {
-    chunk_sizes <- sapply(values, FUN = length)
-    chunk_sizes <- table(chunk_sizes)
-    chunk_summary <- sprintf("%d chunks with %s elements",
-                             chunk_sizes, names(chunk_sizes))
-    chunk_summary <- paste(chunk_summary, collapse = ", ")
-    msg <- sprintf("Unexpected error in doFuture(): After gathering and merging the values from %d chunks in to a list, the total number of elements (= %d) does not match the number of input elements in 'X' (= %d). There were in total %d chunks and %d elements (%s)", nchunks, length(values2), nX, nchunks, sum(chunk_sizes), chunk_summary)
-    if (debug) {
-      mdebug(msg)
-      message(capture.output(print(chunk_sizes)))
-      mdebug("Results before merge chunks:")
-      message(capture.output(str(values)))
-      mdebug("Results after merge chunks:")
-      message(capture.output(str(values2)))
-    }
-    msg <- sprintf("%s. Example of the first few values: %s", msg,
-                   paste(capture.output(str(head(values2, 3L))),
-                         collapse = "\\n"))
-    ex <- FutureError(msg)
-    stop(ex)
-  }
+  assert_values2(nX, values, values2, fcn = "future_lapply()", debug = debug)
   values <- values2
   rm(list = "values2")
   
